#include "print.h"
#include <suifkernel/suifkernel_messages.h>

static LString sensorAnnote("sensor");
static LString localAnnote("local");
static LString constantQualifier("const");

EsterelPrinter::EsterelPrinter(SuifEnv *env)
{
  map = new VisitorMap(env);

#define VISITOR(class) \
 map->register_visit_method(this, &visit ## class, class ::get_class_name());

  VISITORS;
#undef VISITOR

  map->register_unknown_method( this, &visitUnknown);

  _operatorPrecedence.enter_value("or",1);
  _operatorPrecedence.enter_value("and",2);
  _operatorPrecedence.enter_value("not",3);

  _operatorPrecedence.enter_value("=",4);
  _operatorPrecedence.enter_value("<>",4);
  _operatorPrecedence.enter_value("<",4);
  _operatorPrecedence.enter_value(">",4);
  _operatorPrecedence.enter_value("<=",4);
  _operatorPrecedence.enter_value(">=",4);

  _operatorPrecedence.enter_value("+",5);
  _operatorPrecedence.enter_value("-",5);

  _operatorPrecedence.enter_value("*",6);
  _operatorPrecedence.enter_value("/",6);
  _operatorPrecedence.enter_value("mod",6);

  _operatorPrecedence.enter_value("?",7);
  _operatorPrecedence.enter_value("pre",7);
  _operatorPrecedence.enter_value("??",7);
 
}

EsterelPrinter::~EsterelPrinter()
{
  delete map;
}

void EsterelPrinter::State::add_indent(String &result)
{
  result += "\n";
  for (int i = get_indent() ; --i >= 0 ; ) result += "  "; 
}

void EsterelPrinter::expression(Expression *e)
{
  suif_assert(e);
  State st(state);
  st.indent();
  st.set_precedence(0);
  state = &st;
  map->apply(e);
  pop_state();
}

void EsterelPrinter::sigExpression(Expression *e)
{
  suif_assert(e);
  bool needBrackets =
    is_a<BinaryExpression>(e) || is_a<UnaryExpression>(e);
  if (needBrackets) result += "[ ";
  State st(state);
  st.indent();
  st.set_precedence(0);
  state = &st;
  map->apply(e);
  if (needBrackets) result += " ]";
  pop_state();
}

void EsterelPrinter::statement(Statement *s)
{
  suif_assert(s);
  State st(state);
  st.indent();
  st.add_indent(result);

  state = &st;
  map->apply(s);
  pop_state();

  indent();
}

void EsterelPrinter::fileSetBlock(const FileSetBlock *fsb)
{
  suif_assert(fsb != 0);
  for ( Iter<FileBlock*> fbiter = fsb->get_file_block_iterator() ;
	fbiter.is_valid() ; fbiter.next() ) {

    DefinitionBlock *db = fbiter.current()->get_definition_block();
    suif_assert( db );

    for ( Iter<ProcedureDefinition*> pditer =
	    db->get_procedure_definition_iterator() ;
	  pditer.is_valid() ; pditer.next() ) {
      suif_assert_message( is_a<ModuleDefinition>(pditer.current()),
			   ("non-module definition in procedure definition block") );
      ModuleDefinition *md = to<ModuleDefinition>(pditer.current());
      module(md);
      result += "\n\n";
    }
  }
}

void EsterelPrinter::module(const ModuleDefinition* md)
{
  suif_assert(md);

  State st;
  st.indent();
  state = &st;

  result += "module ";
  suif_assert( md->get_procedure_symbol() != 0 );
  suif_assert_message( is_a<ModuleSymbol>(md->get_procedure_symbol()),
	       ("Should have a module symbol attached to a module definition"));  
  result += md->get_procedure_symbol()->get_name();
  result += ":\n";

  types(md);
  constants(md);
  functions(md);
  procedures(md);
  tasks(md);
  signals(md);
  relations(md);

  result += "\n";

  map->apply(md->get_body());

  result += "\n\n";

  result += "end module";
}

void EsterelPrinter::types(const ModuleDefinition *md)
{
  suif_assert(md);
  suif_assert(md->get_type_symbols());
  for ( Iter<SymbolTableObject *> iter =
	  md->get_type_symbols()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    if ( is_a<QualifiedType>(iter.current()) ) {
      QualifiedType *qt = to<QualifiedType>(iter.current());
      if ( is_a<ExternalType>(qt->get_base_type()) ) {
	result += "  type ";
	result += qt->get_name();
	result += ";\n";
      }
    }
  }
}

void EsterelPrinter::constants(const ModuleDefinition *md)
{
  suif_assert(md);
  suif_assert(md->get_symbol_table());
  for ( Iter<SymbolTableObject *> iter =
	  md->get_symbol_table()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    if ( is_a<VariableSymbol>(iter.current()) ) {
      VariableSymbol *vs = to<VariableSymbol>(iter.current());     
      QualifiedType *qt = vs->get_type();
      if (qt->has_qualification_member(constantQualifier)) {
	result += "  constant ";
	result += vs->get_name();
	suif_assert(vs->get_definition());
	ValueBlock *vb = vs->get_definition()->get_initialization();
	if (vb) {
	  ExpressionValueBlock *evb = to<ExpressionValueBlock>(vb);
	  result += " = ";
	  expression(evb->get_expression());
	}
	result += " : ";
	result += qt->get_name();
	result += ";\n";
      }
    }
  }
}

void EsterelPrinter::functions(const ModuleDefinition *md)
{
  suif_assert(md);
  suif_assert(md->get_function_symbols());
  for ( Iter<SymbolTableObject *> iter =
	  md->get_function_symbols()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    if ( is_a<EsterelFunctionSymbol>(iter.current()) ) {
      EsterelFunctionSymbol *efs = to<EsterelFunctionSymbol>(iter.current());
      EsterelFunctionType *eft = efs->get_type();
      result += "  function ";
      result += efs->get_name();
      result += "(";
      Iter<QualifiedType*> titer = eft->get_argument_iterator();
      while (titer.is_valid()) {
	result += titer.current()->get_name();
	titer.next();
	if ( titer.is_valid() ) result += ", ";
      }
      result += ") : ";
      result += eft->get_result()->get_name();
      result += ";\n";
    }
  }
}

void EsterelPrinter::procedures(const ModuleDefinition *md)
{
  suif_assert(md);
  suif_assert(md->get_procedure_symbols());
  for ( Iter<SymbolTableObject *> iter =
	  md->get_procedure_symbols()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    if ( is_a<EsterelProcedureSymbol>(iter.current()) ) {
      EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(iter.current());
      EsterelProcedureType *ept = eps->get_type();
      result += "  procedure ";
      result += eps->get_name();
      result += "(";
      Iter<QualifiedType*> titer = ept->get_reference_argument_iterator();
      while (titer.is_valid()) {
	result += titer.current()->get_name();
	titer.next();
	if ( titer.is_valid() ) result += ", ";
      }
      result += ")(";
      titer = ept->get_value_argument_iterator();
      while (titer.is_valid()) {
	result += titer.current()->get_name();
	titer.next();
	if ( titer.is_valid() ) result += ", ";
      }
      result += ");\n";
    }
  }
}

void EsterelPrinter::tasks(const ModuleDefinition *md)
{
  suif_assert(md);
  suif_assert(md->get_task_symbols());
  for ( Iter<SymbolTableObject *> iter =
	  md->get_task_symbols()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    if ( is_a<EsterelProcedureSymbol>(iter.current()) ) {
      EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(iter.current());
      EsterelProcedureType *ept = eps->get_type();
      result += "  task ";
      result += eps->get_name();
      result += "(";
      Iter<QualifiedType*> titer = ept->get_reference_argument_iterator();
      while (titer.is_valid()) {
	result += titer.current()->get_name();
	titer.next();
	if ( titer.is_valid() ) result += ", ";
      }
      result += ")(";
      titer = ept->get_value_argument_iterator();
      while (titer.is_valid()) {
	result += titer.current()->get_name();
	titer.next();
	if ( titer.is_valid() ) result += ", ";
      }
      result += ");\n";
    }
  }
}

void EsterelPrinter::signals(const ModuleDefinition *md)
{
  suif_assert(md);
  suif_assert(md->get_signal_symbols());
  for ( Iter<SymbolTableObject *> iter =
	  md->get_signal_symbols()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    if ( is_a<SignalSymbol>(iter.current()) ) {
      SignalSymbol *ss = to<SignalSymbol>(iter.current());
      result += "  ";
      SignalType *st = to<SignalType>(ss->get_type());
      result += st->get_direction();
      suif_assert_message( !(st->get_direction() == sensorAnnote)
			   || st->get_base_type(), ("sensor `%s' has no type",
					            st->get_name().c_str() ) );
      result += " ";
      result += ss->get_name();
      suif_assert(ss->get_definition());
      ValueBlock *vb = ss->get_definition()->get_initialization();
      if ( vb ) {
	ExpressionValueBlock *evb = to<ExpressionValueBlock>(vb);
	result += " := ";
	expression(evb->get_expression());
      }
      if ( !is_a<VoidType>(st->get_base_type()) ) {
	result += " : ";
	if (st->get_combine()) result += "combine ";
	result += st->get_base_type()->get_name();
	if (st->get_combine()) {
	  result += " with ";
	  result += st->get_combine()->get_name();
	}
      }
      result += ";\n";
    }
  }
}

void EsterelPrinter::relations(const ModuleDefinition *md)
{
  suif_assert(md);
  if ( md->get_relation_count() ) {
    result += "  relation ";
    Iter<InputRelation *> iter = md->get_relation_iterator() ;
    while (iter.is_valid()) {
      if (is_a<ImpliesRelation>(iter.current())) {
	ImpliesRelation *ir = to<ImpliesRelation>(iter.current());
	suif_assert(ir->get_predicate());
	result += ir->get_predicate()->get_name();
	result += " => ";
	suif_assert(ir->get_implication());
	result += ir->get_implication()->get_name();
      } else {
	ExclusionRelation *er = to<ExclusionRelation>(iter.current());
	Iter<SignalSymbol *> ssiter =  er->get_signal_iterator();
	while (ssiter.is_valid()) {
	  result += ssiter.current()->get_name();
	  ssiter.next();
	  if (ssiter.is_valid()) result += " # ";
	}
      }
      iter.next();
      if (iter.is_valid()) {
	result += ",";
	indent();
	result += "         ";
      }
    }
    result +=";\n";
  }
}

int EsterelPrinter::op_precedence(LString op)
{
  suif_map<LString, int>::iterator iter = _operatorPrecedence.find(op);
  suif_assert_message( iter != _operatorPrecedence.end(),
		       ("precedence not defined for `%s'", op.c_str()));
  return (*iter).second;
}

#define VISITOR(class) \
void EsterelPrinter::print ## class(const class* stmt)

VISITOR(NothingStatement)
{
  result += "nothing";
}

VISITOR(PauseStatement)
{
  result += "pause";
}

VISITOR(HaltStatement)
{
  result += "halt";
}

VISITOR(StatementList)
{
  State st(state);
  st.set_precedence(sequentialPrecedence);
  state = &st;
  Iter<Statement *>iter = stmt->get_statement_iterator();
  while ( iter.is_valid() ) {
    map->apply(iter.current());
    iter.next();
    if ( iter.is_valid() ) {
      result += ";";
      indent();
    }
  }
  pop_state();
}

VISITOR(ParallelStatementList)
{
  State st(state);
  st.indent();
  bool needBrackets = st.set_precedence(parallelPrecedence);

  if (needBrackets) {
    result += "[";
    indent();
  }
  Iter<Statement *>iter = stmt->get_statement_iterator();
  while ( iter.is_valid() ) {
    result += "  ";
    state = &st;
    map->apply(iter.current());
    pop_state();
    iter.next();
    if ( iter.is_valid() ) {
      indent();
      result += "||";
      indent();
    }
  }
  if (needBrackets) {
    indent();
    result += "]";
  }
}

VISITOR(StoreVariableStatement)
{
  suif_assert(stmt->get_destination());
  suif_assert(stmt->get_value());
  result += stmt->get_destination()->get_name();
  result += " := ";

  expression(stmt->get_value());
}

VISITOR(LoadVariableExpression)
{
  suif_assert(stmt->get_source());
  result += stmt->get_source()->get_name();
}

VISITOR(LoadTrapExpression)
{
  suif_assert(stmt->get_source());
  result += stmt->get_source()->get_name();
}

VISITOR(LoadTrapValueExpression)
{
  suif_assert(stmt->get_source());
  result += "??";
  result += stmt->get_source()->get_name();
}

VISITOR(LoadSignalExpression)
{
  suif_assert(stmt->get_source());
  result += stmt->get_source()->get_name();
}

VISITOR(LoadSignalValueExpression)
{
  suif_assert(stmt->get_source());
  result += "?";
  result += stmt->get_source()->get_name();
}

VISITOR(Delay)
{
  Expression *p = stmt->get_predicate();
  Expression *c = stmt->get_count();
  if (stmt->get_is_immediate()) {
    suif_assert(c == 0);
    result += "immediate ";
    suif_assert(p);
    sigExpression(p);
  } else {
    suif_assert(c);
    suif_assert(p);
    expression(c);
    result += " ";
    sigExpression(p);
  }
}

VISITOR(IntConstant)
{
  if (is_a<BooleanType>(stmt->get_result_type()) ) {
    if ( stmt->get_value() == 0 ) result += "false";
    else result += "true";
  } else {
      char * constant = stmt->get_value().to_string();
      result += constant;
      delete [] constant;
  }
}

VISITOR(StringConstant)
{
  result += "\"";
  const char * c = stmt->get_value().c_str();
  for (const char *p = c ; *p != '\0' ; p++) {
    if (*p == '\"')
      result += "\"";
    result += *p;
  }
  result += "\"";
}

VISITOR(FloatConstant)
{
  result += stmt->get_value();
}

VISITOR(BinaryExpression)
{
  suif_assert(stmt->get_source1());
  suif_assert(stmt->get_source2());
  State st(state);

  bool needParen = st.set_precedence(op_precedence(stmt->get_opcode()));

  if ( needParen ) result += "(";


  state = &st;
  map->apply(stmt->get_source1());

  result += " ";
  result += stmt->get_opcode();
  result += " ";

  map->apply(stmt->get_source2());

  if (needParen) result += ")";
  pop_state();
}

VISITOR(UnaryExpression)
{
  suif_assert(stmt->get_source());
  State st(state);
  state = &st;
  bool needParen = st.set_precedence(op_precedence(stmt->get_opcode()));

  result += stmt->get_opcode();
  result += " ";
  if ( needParen ) result += "(";
  map->apply(stmt->get_source());
  if (needParen) result += ")";
  pop_state();
}

VISITOR(VariableStatement)
{
  result += "var ";
  Iter<SymbolTableObject*>iter =
    stmt->get_symbol_table()->get_symbol_table_object_iterator();
  while (iter.is_valid()) {
    VariableSymbol *vs = to<VariableSymbol>(iter.current());
    result += vs->get_name();
    VariableDefinition *vd = vs->get_definition();
    suif_assert(vd);
    if ( vd->get_initialization() ) {
      ExpressionValueBlock *evb =
	to<ExpressionValueBlock>(vd->get_initialization());
      result += " := ";
      expression(evb->get_expression());
    }
    result += " : ";
    result += vs->get_type()->get_name();
    iter.next();
    if ( iter.is_valid() ) {
      result += ",";
      indent();
      result += "    ";
    }
  }
  result += " in";
  statement(stmt->get_body());
  result += "end var";
}

VISITOR(TrapStatement)
{
  result += "trap ";
  Iter<SymbolTableObject*>iter =
    stmt->get_symbol_table()->get_symbol_table_object_iterator();
  while (iter.is_valid()) {
    VariableSymbol *vs = to<VariableSymbol>(iter.current());
    result += vs->get_name();
    VariableDefinition *vd = vs->get_definition();
    suif_assert(vd);
    if ( vd->get_initialization() ) {
      ExpressionValueBlock *evb =
	to<ExpressionValueBlock>(vd->get_initialization());
      result += " := ";
      expression(evb->get_expression());
    }
    suif_assert(vs->get_type());
    suif_assert(vs->get_type()->get_base_type());
    if ( !is_a<VoidType>(vs->get_type()->get_base_type()) ) {
      result += " : ";
      result += vs->get_type()->get_name();
    }
    iter.next();
    if ( iter.is_valid() ) {
      result += ",";
      indent();
      result += "    ";
    }
  }
  result += " in";
  statement(stmt->get_body());
  for ( Iter<PredicatedStatement*> psi = stmt->get_handler_iterator() ;
	psi.is_valid() ; psi.next() ) {
    result += "handle ";
    expression(psi.current()->get_predicate());
    result += " do";
    statement(psi.current()->get_body());
  }
  result += "end trap";
}

VISITOR(SignalStatement)
{
  result += "signal ";
  Iter<SymbolTableObject*>iter =
    stmt->get_symbol_table()->get_symbol_table_object_iterator();
  while (iter.is_valid()) {
    SignalSymbol *ss = to<SignalSymbol>(iter.current());
    result += ss->get_name();
    VariableDefinition *vd = ss->get_definition();
    suif_assert(vd);
    if ( vd->get_initialization() ) {
      ExpressionValueBlock *evb =
	to<ExpressionValueBlock>(vd->get_initialization());
      result += " := ";
      expression(evb->get_expression());
    }
    suif_assert(ss->get_type());
    suif_assert(is_a<SignalType>(ss->get_type()));
    suif_assert(to<SignalType>(ss->get_type())->get_direction()
		== localAnnote );
    suif_assert(ss->get_type()->get_base_type());
    if ( !is_a<VoidType>(ss->get_type()->get_base_type()) ) {
      result += " : ";
      result += ss->get_type()->get_base_type()->get_name();
    }
    iter.next();
    if ( iter.is_valid() ) {
      result += ",";
      indent();
      result += "    ";
    }
  }
  result += " in";
  statement(stmt->get_body());
  result += "end signal";
}

VISITOR(EmitStatement)
{
  suif_assert(stmt->get_destination());
  result += "emit ";
  result += stmt->get_destination()->get_name();
  if ( stmt->get_value() ) {
    result += "(";
    expression(stmt->get_value());
    result += ")";
  }
}

VISITOR(ExitStatement)
{
  result += "exit ";
  result += stmt->get_trap()->get_name();
  if ( stmt->get_value() ) {
    result += "(";
    expression(stmt->get_value());
    result += ")";
  }
}

VISITOR(SustainStatement)
{
  suif_assert(stmt->get_destination());
  result += "sustain ";
  result += stmt->get_destination()->get_name();
  if ( stmt->get_value() ) {
    result += "(";
    expression(stmt->get_value());
    result += ")";
  }
}

VISITOR(PresentStatement)
{
  result += "present ";
  if ( stmt->get_case_count() == 1 ) {
    PredicatedStatement *ps = stmt->get_case(0);
    suif_assert(ps->get_predicate());
    sigExpression(ps->get_predicate());
    result += " ";
    if ( ps->get_body() ) {
      result += "then";
      statement(ps->get_body());
    }
  } else {
    State st(state);
    st.indent();
    state = &st;
    indent();
    for ( Iter<PredicatedStatement*> iter = stmt->get_case_iterator() ;
	  iter.is_valid() ; iter.next() ) {
      PredicatedStatement *ps = iter.current();
      result += "case ";
      suif_assert(ps->get_predicate());
      sigExpression(ps->get_predicate());
      if (ps->get_body()) {
	result += " do";
	statement(ps->get_body());
      } else {
	indent();
      }
    }
    pop_state();
  }
  if ( stmt->get_default_case() ) {
    result += "else";
    statement(stmt->get_default_case());
  }
  result += "end present";
}

VISITOR(EsterelIfStatement)
{
  result += "if ";
  suif_assert( stmt->get_case_count() >= 1);
  PredicatedStatement *ps = stmt->get_case(0);
  suif_assert(ps->get_predicate());
  expression(ps->get_predicate());
  result += " ";
  if (ps->get_body()) {
    result += "then";
    statement(ps->get_body());
  }
  Iter<PredicatedStatement *> iter = stmt->get_case_iterator();
  iter.next();
  while (iter.is_valid()) {
    result += "elsif ";
    Expression *e = iter.current()->get_predicate();
    suif_assert(e);
    expression(e);
    Statement *s = iter.current()->get_body();
    suif_assert(s);
    result += " then";
    statement(s);
    iter.next();
  }
  if ( stmt->get_default_case() ) {
    result += "else";
    statement(stmt->get_default_case());
  }
  result += "end if";
}

VISITOR(LoopStatement)
{
  result += "loop";
  statement(stmt->get_body());
  result += "end loop";
}

VISITOR(LoopEachStatement)
{
  result += "loop";
  statement(stmt->get_body());
  result += "each ";
  sigExpression(stmt->get_predicate());
}

VISITOR(EveryStatement)
{
  result += "every ";
  sigExpression(stmt->get_predicate());
  result += " do";
  statement(stmt->get_body());
  result += "end every";
}

VISITOR(SuspendStatement)
{
  result += "suspend";
  statement(stmt->get_body());
  result += "when ";
  sigExpression(stmt->get_predicate());
}

VISITOR(DoWatchingStatement)
{
  result += "do";
  statement(stmt->get_body());
  result += "watching ";
  sigExpression(stmt->get_predicate());
  if (stmt->get_timeout()) {
    result += " timeout";
    statement(stmt->get_timeout());
    result += "end timeout";
  }
}

VISITOR(DoUptoStatement)
{
  result += "do";
  statement(stmt->get_body());
  result += "upto ";
  sigExpression(stmt->get_predicate());
}

VISITOR(RepeatStatement)
{
  if (stmt->get_positive()) {
    result += "positive ";
  }
  result += "repeat ";
  expression(stmt->get_count());
  result += " times";
  statement(stmt->get_body());
  result += "end repeat";
}

VISITOR(AbortStatement)
{
  if (stmt->get_weak()) result += "weak ";
  result += "abort";
  statement(stmt->get_body());
  result += "when";

  if ( stmt->get_case_count() == 1 ) {
    PredicatedStatement *ps = stmt->get_case(0);
    result += " ";
    sigExpression(ps->get_predicate());
    if (ps->get_body()) {
      result += " do";
      statement(ps->get_body());
      goto PrintEnd;
    }
  } else {
    indent();
    for ( Iter<PredicatedStatement*> iter = stmt->get_case_iterator() ;
	  iter.is_valid() ; iter.next() ) {
      PredicatedStatement *ps = iter.current();
      result += "case ";
      sigExpression(ps->get_predicate());
      if (ps->get_body()) {
	result += " do";
	statement(ps->get_body());
      } else {
	indent();
      }
    }
  PrintEnd:
    result += "end ";
    if ( stmt->get_weak()) result += "weak ";
    result += "abort";    
  }
}


VISITOR(AwaitStatement)
{
  result += "await ";
  if ( stmt->get_case_count() == 1 ) {
    PredicatedStatement *ps = stmt->get_case(0);
    sigExpression(ps->get_predicate());
    if (ps->get_body()) {
      result += " do";
      statement(ps->get_body());
      result += "end await";
    }
  } else {
    indent();
    for ( Iter<PredicatedStatement*> iter = stmt->get_case_iterator() ;
	  iter.is_valid() ; iter.next() ) {
      PredicatedStatement *ps = iter.current();
      result += "case ";
      sigExpression(ps->get_predicate());
      if (ps->get_body()) {
	result += " do";
	statement(ps->get_body());
      } else {
	indent();
      }
    }
    result += "end await";
  }
}

VISITOR(ProcedureCallStatement)
{
  suif_assert(stmt->get_procedure());
  result += "call ";
  result += stmt->get_procedure()->get_name();
  result += "(";
  Iter<VariableSymbol *> vsi = stmt->get_reference_arg_iterator();
  while (vsi.is_valid()) {
    result += vsi.current()->get_name();    
    vsi.next();
    if (vsi.is_valid()) result += ", ";
  }
  result += ")(";
  Iter<Expression *> ei = stmt->get_value_arg_iterator();
  while (ei.is_valid()) {
    expression(ei.current());
    ei.next();
    if (ei.is_valid()) result += ", ";
  }
  result += ")";
}

VISITOR(ExecStatement)
{
  result += "exec ";
  suif_assert(stmt->get_call_count() >= 1);
  TaskCall *tc;
  if (stmt->get_call_count() == 1) {	
    tc = stmt->get_call(0);
    taskCall(tc);
    if (tc->get_body()) result += "end exec";
  } else {
    indent();
    for ( Iter<TaskCall *>iter = stmt->get_call_iterator() ;
	  iter.is_valid(); iter.next()) {
      result += "case ";
      tc = iter.current();
      taskCall(iter.current());
      if (!tc->get_body()) indent();
    }
  result += "end exec";
  }
}

VISITOR(FunctionCall)
{
  suif_assert(stmt->get_callee());
  result += stmt->get_callee()->get_name();
  result += "(";
  Iter<Expression *> iter = stmt->get_argument_iterator();
  while (iter.is_valid()) {
    expression(iter.current());
    iter.next();
    if (iter.is_valid()) result += ", ";
  }
  result += ")";
}

VISITOR(RunStatement)
{
  result += "run ";
  if ( stmt->get_new_module_name() != stmt->get_old_module_name() ) {
    result += stmt->get_new_module_name();
    result += " / ";
  }
  result += stmt->get_old_module_name();

  if (stmt->get_type_renaming_count() ||
      stmt->get_constant_renaming_count() ||
      stmt->get_function_renaming_count() ||
      stmt->get_procedure_renaming_count() ||
      stmt->get_task_renaming_count() ||
      stmt->get_signal_renaming_count() ) {

    result += " [";
    State st(state);
    st.indent();
    state = &st;

    indent();

    bool more = false;

    if (stmt->get_type_renaming_count()) {
      result += "type ";
      Iter<TypeRenaming *> iter = stmt->get_type_renaming_iterator();
      while (iter.is_valid()) {
	result += iter.current()->get_new_type()->get_name();
	result += " / ";
	result += iter.current()->get_old_name();
	iter.next();
	if (iter.is_valid()) {
	  result += ",";
	  indent();
	  result += "     ";
	}
      }
      more = true;
    }

    if (stmt->get_constant_renaming_count()) {
      if (more) {
	result += ";";
	indent();
      }
      result += "constant ";
      Iter<ConstantRenaming *> iter = stmt->get_constant_renaming_iterator();
      while (iter.is_valid()) {
	Expression *e = iter.current()->get_new_value();
	suif_assert(e);
	expression(e);
	result += " / ";
	result += iter.current()->get_old_name();
	iter.next();
	if (iter.is_valid()) {
	  result += ",";
	  indent();
	  result += "         ";
	}
      }
      more = true;      
    }

    if (stmt->get_function_renaming_count()) {
      if (more) {
	result += ";";
	indent();
      }
      result += "function ";
      Iter<FunctionRenaming *> iter = stmt->get_function_renaming_iterator();
      while (iter.is_valid()) {
	EsterelFunctionSymbol *efs = iter.current()->get_new_function();
	suif_assert(efs);
	result += efs->get_name();
	result += " / ";
	result += iter.current()->get_old_name();
	iter.next();
	if (iter.is_valid()) {
	  result += ",";
	  indent();
	  result += "         ";
	}
      }
      more = true;      
    }

    if (stmt->get_procedure_renaming_count()) {
      if (more) {
	result += ";";
	indent();
      }
      result += "procedure ";
      Iter<ProcedureRenaming *> iter = stmt->get_procedure_renaming_iterator();
      while (iter.is_valid()) {
	EsterelProcedureSymbol *eps = iter.current()->get_new_procedure();
	suif_assert(eps);
	result += eps->get_name();
	result += " / ";
	result += iter.current()->get_old_name();
	iter.next();
	if (iter.is_valid()) {
	  result += ",";
	  indent();
	  result += "          ";
	}
      }
      more = true;      
    }

    if (stmt->get_task_renaming_count()) {
      if (more) {
	result += ";";
	indent();
      }
      result += "task ";
      Iter<ProcedureRenaming *> iter = stmt->get_task_renaming_iterator();
      while (iter.is_valid()) {
	EsterelProcedureSymbol *eps = iter.current()->get_new_procedure();
	suif_assert(eps);
	result += eps->get_name();
	result += " / ";
	result += iter.current()->get_old_name();
	iter.next();
	if (iter.is_valid()) {
	  result += ",";
	  indent();
	  result += "     ";
	}
      }
      more = true;      
    }

    if (stmt->get_signal_renaming_count()) {
      if (more) {
	result += ";";
	indent();
      }
      result += "signal ";
      Iter<SignalRenaming *> iter = stmt->get_signal_renaming_iterator();
      while (iter.is_valid()) {
	SignalSymbol *ss = iter.current()->get_new_signal();
	suif_assert(ss);
	result += ss->get_name();
	result += " / ";
	result += iter.current()->get_old_name();
	iter.next();
	if (iter.is_valid()) {
	  result += ",";
	  indent();
	  result += "       ";
	}
      }
    }
    
    result += " ]";

    pop_state();

  }
}

#undef VISITOR

void EsterelPrinter::taskCall(const TaskCall *tc)
{
  suif_assert(tc);
  suif_assert(tc->get_procedure());
  result += tc->get_procedure()->get_name();
  result += "(";
  Iter<VariableSymbol *> vsi = tc->get_reference_arg_iterator();
  while (vsi.is_valid()) {
    result += vsi.current()->get_name();    
    vsi.next();
    if (vsi.is_valid()) result += ", ";
  }
  result += ")(";
  Iter<Expression *> ei = tc->get_value_arg_iterator();
  while (ei.is_valid()) {
    expression(ei.current());
    ei.next();
    if (ei.is_valid()) result += ", ";
  }
  result += ") return ";
  suif_assert(tc->get_signal());  
  result += tc->get_signal()->get_name();
  result += " ";
  if (tc->get_body()) {
    result += "do ";
    statement(tc->get_body());
  }
}

void EsterelPrinter::printUnknown(Object *obj)
{
  result += "<";
  result += obj->getClassName();
  result += "> ";
}

